import dataclasses

import pytest
import skhep_testdata

from pylhe import (
    LHEEventInfo,
    LHEFile,
    LHEInit,
    LHEInitInfo,
    LHEParticle,
    LHEProcInfo,
    read_lhe,
)

TEST_FILE = skhep_testdata.data_path("pylhe-testfile-pr29.lhe")


def test_LHEEvent():
    events = read_lhe(TEST_FILE)
    event = next(events)  # it contains 8 pions and a proton

    assert event.eventinfo is not None

    assert len(event.particles) == 9

    for p in event.particles:
        assert p.event == event

    assert event._graph is None


def test_LHEEventInfo_no_default_init():
    with pytest.raises(TypeError):
        _ = LHEEventInfo()


def test_LHEEventInfo_fromstring():
    """
    Data taken from the first <event> block of scikit-hep-testdata's package
    "pylhe-testlhef3.lhe" file.
    """
    data = "5 66 0.50109093E+02 0.14137688E+03 0.75563862E-02 0.12114027E+00"
    event_info = LHEEventInfo.fromstring(data)

    assert event_info.nparticles == 5
    assert event_info.pid == 66
    assert event_info.weight == pytest.approx(0.50109093e02)
    assert event_info.scale == pytest.approx(0.14137688e03)
    assert event_info.aqed == pytest.approx(0.75563862e-02)
    assert event_info.aqcd == pytest.approx(0.12114027e00)


def test_LHEEventInfo_backwards_compatibility():
    """
    Test backwards-compatibility of fieldnames.
    """
    event_info = LHEEventInfo(
        nparticles=6, pid=67, weight=0.6, scale=0.2, aqed=0.8, aqcd=0.2
    )

    assert event_info.fieldnames == [
        "nparticles",
        "pid",
        "weight",
        "scale",
        "aqed",
        "aqcd",
    ]


def test_LHEFile_no_default_init():
    with pytest.raises(TypeError):
        _ = LHEFile()


def test_LHEInit_no_default_init():
    with pytest.raises(TypeError):
        _ = LHEInit()


def test_LHEInit_fromstring():
    """
    Data taken from the <init> block of scikit-hep-testdata's package
    "pylhe-testlhef3.lhe" file.
    """
    data = "2212 2212 0.40000000E+04 0.40000000E+04 -1 -1 21100 21100 -4 1"
    result = {
        "beamA": 2212.0,
        "beamB": 2212.0,
        "energyA": 4000.0,
        "energyB": 4000.0,
        "PDFgroupA": -1.0,
        "PDFgroupB": -1.0,
        "PDFsetA": 21100.0,
        "PDFsetB": 21100.0,
        "weightingStrategy": -4.0,
        "numProcesses": 1.0,
    }
    assert dataclasses.asdict(LHEInitInfo.fromstring(data)) == result


def test_LHEParticle_no_default_init():
    with pytest.raises(TypeError):
        _ = LHEParticle()


def test_LHEParticle_fromstring():
    """
    Data taken from the first <event> block of scikit-hep-testdata's package
    "pylhe-testlhef3.lhe" file.
    """
    particles = [
        "        5 -1    0    0  501    0 0.00000000E+00 0.00000000E+00 0.14322906E+03 0.14330946E+03 0.48000000E+01 0.0000E+00 0.0000E+00",
        "        2 -1    0    0  502    0 0.00000000E+00 0.00000000E+00 -.93544317E+03 0.93544323E+03 0.33000000E+00 0.0000E+00 0.0000E+00",
        "       24  1    1    2    0    0 -.84258804E+02 -.15708566E+03 -.10629600E+03 0.22257162E+03 0.80398000E+02 0.0000E+00 0.0000E+00",
        "        5  1    1    2  501    0 -.13668073E+03 -.36307424E+02 -.40614473E+02 0.14721558E+03 0.48000000E+01 0.0000E+00 0.0000E+00",
        "        1  1    1    2  502    0 0.22093954E+03 0.19339308E+03 -.64530364E+03 0.70896548E+03 0.33000000E+00 0.0000E+00 0.0000E+00",
    ]

    particle_objs = [LHEParticle.fromstring(p) for p in particles]

    assert [p.id for p in particle_objs] == [5, 2, 24, 5, 1]
    assert [p.status for p in particle_objs] == [-1.0, -1.0, 1.0, 1.0, 1.0]
    assert [p.mother1 for p in particle_objs] == [0.0, 0.0, 1.0, 1.0, 1.0]
    assert [p.mother2 for p in particle_objs] == [0.0, 0.0, 2.0, 2.0, 2.0]
    assert [p.color1 for p in particle_objs] == [501.0, 502.0, 0.0, 501.0, 502.0]
    assert [p.color2 for p in particle_objs] == [0.0, 0.0, 0.0, 0.0, 0.0]
    assert [p.px for p in particle_objs] == [
        0.0,
        0.0,
        -84.258804,
        -136.68073,
        220.93954,
    ]
    assert [p.py for p in particle_objs] == [
        0.0,
        0.0,
        -157.08566,
        -36.307424,
        193.39308,
    ]
    assert [p.pz for p in particle_objs] == [
        143.22906,
        -935.44317,
        -106.296,
        -40.614473,
        -645.30364,
    ]
    assert [p.e for p in particle_objs] == [
        143.30946,
        935.44323,
        222.57162,
        147.21558,
        708.96548,
    ]
    assert [p.m for p in particle_objs] == [4.8, 0.33, 80.398, 4.8, 0.33]
    assert [p.lifetime for p in particle_objs] == [0.0, 0.0, 0.0, 0.0, 0.0]
    assert [p.spin for p in particle_objs] == [0.0, 0.0, 0.0, 0.0, 0.0]


def test_LHEParticle_backwards_compatibility():
    """
    Test backwards-compatibility of fieldnames.
    """
    particle = LHEParticle(
        id=5,
        status=-1,
        mother1=0,
        mother2=0,
        color1=501,
        color2=0,
        px=0,
        py=0,
        pz=143.22906,
        e=143.30946,
        m=4.8,
        lifetime=0,
        spin=0,
    )

    assert particle.fieldnames == [
        "id",
        "status",
        "mother1",
        "mother2",
        "color1",
        "color2",
        "px",
        "py",
        "pz",
        "e",
        "m",
        "lifetime",
        "spin",
    ]

    # particle is not associated to an event thus mothers should raise a ValueError
    with pytest.raises(ValueError, match=r"Particle is not associated to an event."):
        _ = particle.mothers()


def test_LHEProcInfo_no_default_init():
    with pytest.raises(TypeError):
        _ = LHEProcInfo()


def test_LHEProcInfo_fromstring():
    """
    Data taken from the <init> block of scikit-hep-testdata's package
    "pylhe-testlhef3.lhe" file.
    """
    data = "0.50109086E+02 0.89185414E-01 0.50109093E+02 66"
    result = {
        "xSection": 50.109086,
        "error": 0.089185414,
        "unitWeight": 50.109093,
        "procId": 66.0,
    }
    assert dataclasses.asdict(LHEProcInfo.fromstring(data)) == result


def test_LHEProcInfo_backwards_compatibility():
    """
    Test backwards-compatibility of dict like access and fieldnames.
    """
    proc_info = LHEProcInfo(
        xSection=50.109086, error=0.089185414, unitWeight=50.109093, procId=66.0
    )

    assert proc_info.fieldnames == ["xSection", "error", "unitWeight", "procId"]

    assert proc_info["xSection"] == pytest.approx(50.109086)
    assert proc_info["error"] == pytest.approx(0.089185414)
    assert proc_info["unitWeight"] == pytest.approx(50.109093)
    assert proc_info["procId"] == pytest.approx(66.0)

    proc_info["xSection"] = 60.0
    proc_info["error"] = 0.1
    proc_info["unitWeight"] = 60.0
    proc_info["procId"] = 67.0

    assert proc_info["xSection"] == pytest.approx(60.0)
    assert proc_info["error"] == pytest.approx(0.1)
    assert proc_info["unitWeight"] == pytest.approx(60.0)
    assert proc_info["procId"] == pytest.approx(67.0)


def test_LHEInitInfo_backwards_compatibility():
    """
    Test backwards-compatibility of dict like access and fieldnames.
    """
    lheii = LHEInitInfo(
        beamA=1,
        beamB=2,
        energyA=3.0,
        energyB=4.0,
        PDFgroupA=-1,
        PDFgroupB=-1,
        PDFsetA=21100,
        PDFsetB=21100,
        weightingStrategy=1,
        numProcesses=1,
    )

    assert lheii.fieldnames == [
        "beamA",
        "beamB",
        "energyA",
        "energyB",
        "PDFgroupA",
        "PDFgroupB",
        "PDFsetA",
        "PDFsetB",
        "weightingStrategy",
        "numProcesses",
    ]

    assert lheii["beamA"] == 1
    assert lheii["beamB"] == 2
    assert lheii["energyA"] == 3.0
    assert lheii["energyB"] == 4.0
    assert lheii["PDFgroupA"] == -1
    assert lheii["PDFgroupB"] == -1
    assert lheii["PDFsetA"] == 21100
    assert lheii["PDFsetB"] == 21100
    assert lheii["weightingStrategy"] == 1
    assert lheii["numProcesses"] == 1

    lheii["beamA"] = 5
    lheii["beamB"] = 6
    lheii["energyA"] = 7.0
    lheii["energyB"] = 8.0
    lheii["PDFgroupA"] = -2
    lheii["PDFgroupB"] = -2
    lheii["PDFsetA"] = 21101
    lheii["PDFsetB"] = 21101
    lheii["weightingStrategy"] = 2
    lheii["numProcesses"] = 2

    assert lheii["beamA"] == 5
    assert lheii["beamB"] == 6
    assert lheii["energyA"] == 7.0
    assert lheii["energyB"] == 8.0
    assert lheii["PDFgroupA"] == -2
    assert lheii["PDFgroupB"] == -2
    assert lheii["PDFsetA"] == 21101
    assert lheii["PDFsetB"] == 21101
    assert lheii["weightingStrategy"] == 2
    assert lheii["numProcesses"] == 2



import itertools

import skhep_testdata

import pylhe


def test_LHEEvent_graph_source():
    lhe_file = skhep_testdata.data_path("pylhe-testfile-pr29.lhe")
    events = pylhe.read_lhe_with_attributes(lhe_file)

    # Get the first event
    event = next(events)
    # ... it contains 8 pions and a proton
    # pi unicode character is '&#x03c0;'
    assert event.graph.source.count("&#x03c0;") == 8
    assert "<td>p</td>" in event.graph.source


def test_LHEEvent_graph_source_nonstandard_pdg():
    lhe_file = skhep_testdata.data_path("pylhe-testfile-pr180.lhe")
    events = pylhe.read_lhe_with_attributes(lhe_file)

    # Get the first event
    event = next(events)
    # building the graph should succeed even though there is
    # a non-standard PDG ID 1023 and the name in the graph
    # source should just be the ID number itself
    assert event.graph.source.count("<td>1023</td>") == 1


def test_LHEEvent_graph_render():
    lhe_file = skhep_testdata.data_path("pylhe-testfile-pr29.lhe")
    events = pylhe.read_lhe_with_attributes(lhe_file)

    event = next(itertools.islice(events, 1, 2))
    event.graph.render(filename="test_event1", format="pdf", cleanup=True)


def test_mime():
    lhe_file = skhep_testdata.data_path("pylhe-testfile-pr29.lhe")
    events = pylhe.read_lhe_with_attributes(lhe_file)

    event = next(itertools.islice(events, 1, 2))
    assert event._repr_mimebundle_() == event.graph._repr_mimebundle_()



import io
import os
import tempfile
from tempfile import NamedTemporaryFile

import pytest

import pylhe


def test_missing_init_block_error():
    """Test that ValueError is raised when no <init> block is found in LHE file."""
    # Create an invalid LHE file content without an <init> block
    invalid_lhe_content = """<LesHouchesEvents version="1.0">
<event>
  2      0 +1.0000000e+00  9.11884000e+01 -1.00000000e+00 -1.00000000e+00
       21 -1    0    0  501  502 +0.00000000e+00 +0.00000000e+00 +4.56308892e+02 +4.56308892e+02 +0.00000000e+00 0.0000e+00 9.0000e+00
       21 -1    0    0  502  501 -0.00000000e+00 -0.00000000e+00 -2.24036073e+02 +2.24036073e+02 +0.00000000e+00 0.0000e+00 9.0000e+00
</event>
</LesHouchesEvents>"""

    # Test with string buffer
    with pytest.raises(ValueError, match=r"No <init> block found in the LHE file"):
        pylhe.LHEFile.fromstring(invalid_lhe_content)

    # Test with file-like object
    buffer = io.StringIO(invalid_lhe_content)
    with pytest.raises(ValueError, match=r"No <init> block found in the LHE file"):
        pylhe.LHEFile.frombuffer(buffer)


def test_missing_init_block_error_with_file():
    """Test that ValueError is raised when reading a file without <init> block."""
    # Create an invalid LHE file content without an <init> block
    invalid_lhe_content = """<LesHouchesEvents version="1.0">
<event>
  2      0 +1.0000000e+00  9.11884000e+01 -1.00000000e+00 -1.00000000e+00
       21 -1    0    0  501  502 +0.00000000e+00 +0.00000000e+00 +4.56308892e+02 +4.56308892e+02 +0.00000000e+00 0.0000e+00 9.0000e+00
       21 -1    0    0  502  501 -0.00000000e+00 -0.00000000e+00 -2.24036073e+02 +2.24036073e+02 +0.00000000e+00 0.0000e+00 9.0000e+00
</event>
</LesHouchesEvents>"""

    # Create a temporary file with invalid content
    with NamedTemporaryFile(mode="w", suffix=".lhe", delete=False) as tmp_file:
        tmp_file.write(invalid_lhe_content)
        tmp_file_path = tmp_file.name

    try:
        # Test reading the file through read_lhe_init function
        with pytest.raises(ValueError, match=r"No <init> block found in the LHE file"):
            pylhe.read_lhe_init(tmp_file_path)
    finally:
        os.unlink(tmp_file_path)


def test_missing_init_block_error_with_only_events():
    """Test that ValueError is raised when file contains only events without init."""
    # Create LHE content with events but no init block
    events_only_content = """<LesHouchesEvents version="1.0">
<event>
  2      0 +1.0000000e+00  9.11884000e+01 -1.00000000e+00 -1.00000000e+00
       21 -1    0    0  501  502 +0.00000000e+00 +0.00000000e+00 +4.56308892e+02 +4.56308892e+02 +0.00000000e+00 0.0000e+00 9.0000e+00
       21 -1    0    0  502  501 -0.00000000e+00 -0.00000000e+00 -2.24036073e+02 +2.24036073e+02 +0.00000000e+00 0.0000e+00 9.0000e+00
</event>
<event>
  2      0 +1.0000000e+00  9.11884000e+01 -1.00000000e+00 -1.00000000e+00
       21 -1    0    0  501  502 +0.00000000e+00 +0.00000000e+00 +4.56308892e+02 +4.56308892e+02 +0.00000000e+00 0.0000e+00 9.0000e+00
       21 -1    0    0  502  501 -0.00000000e+00 -0.00000000e+00 -2.24036073e+02 +2.24036073e+02 +0.00000000e+00 0.0000e+00 9.0000e+00
</event>
</LesHouchesEvents>"""

    with pytest.raises(ValueError, match=r"No <init> block found in the LHE file"):
        pylhe.LHEFile.fromstring(events_only_content)


def test_dataclass_delete_field_error():
    """Test that TypeError is raised when attempting to delete a dataclass field."""
    # Create a simple LHEEventInfo instance to test deletion
    eventinfo = pylhe.LHEEventInfo(
        nparticles=2, pid=1, weight=1.0, scale=100.0, aqed=0.007, aqcd=0.1
    )

    # Test that attempting to delete a field raises TypeError
    with pytest.raises(
        TypeError, match=r"Cannot delete field 'nparticles' from dataclass instance"
    ):
        del eventinfo["nparticles"]

    # Test with a different field
    with pytest.raises(
        TypeError, match=r"Cannot delete field 'weight' from dataclass instance"
    ):
        del eventinfo["weight"]

    # Test with a non-existent field (should also raise TypeError)
    with pytest.raises(
        TypeError, match=r"Cannot delete field 'nonexistent' from dataclass instance"
    ):
        del eventinfo["nonexistent"]


def test_empty_init_block_error():
    """Test that ValueError is raised when <init> block has no text content."""
    # Create LHE content with an empty <init> block
    empty_init_content = """<LesHouchesEvents version="1.0">
<init></init>
<event>
  2      0 +1.0000000e+00  9.11884000e+01 -1.00000000e+00 -1.00000000e+00
       21 -1    0    0  501  502 +0.00000000e+00 +0.00000000e+00 +4.56308892e+02 +4.56308892e+02 +0.00000000e+00 0.0000e+00 9.0000e+00
       21 -1    0    0  502  501 -0.00000000e+00 -0.00000000e+00 -2.24036073e+02 +2.24036073e+02 +0.00000000e+00 0.0000e+00 9.0000e+00
</event>
</LesHouchesEvents>"""

    # Test with string buffer
    with pytest.raises(ValueError, match=r"<init> block has no text"):
        pylhe.LHEFile.fromstring(empty_init_content)

    # Test with file-like object
    buffer = io.StringIO(empty_init_content)
    with pytest.raises(ValueError, match=r"<init> block has no text"):
        pylhe.LHEFile.frombuffer(buffer)


def test_empty_event_block_error():
    """Test that ValueError is raised when <event> block has no text content."""
    # Create LHE content with valid init but empty event block
    empty_event_content = """<LesHouchesEvents version="1.0">
<init>
  2212  2212  6.500000e+03  6.500000e+03  0  0  0  0  3  1
  1.000000e+00  0.000000e+00  1.000000e+00  1
</init>
<event></event>
</LesHouchesEvents>"""

    # Test reading events with empty event block
    with tempfile.NamedTemporaryFile(mode="w", suffix=".lhe", delete=False) as tmp_file:
        tmp_file.write(empty_event_content)
        tmp_file_path = tmp_file.name

    try:
        with pytest.raises(ValueError, match=r"<event> block has no text"):
            list(pylhe.read_lhe(tmp_file_path))
    finally:
        os.unlink(tmp_file_path)


def test_empty_weights_block_error():
    """Test that ValueError is raised when <weights> block has no text content."""
    # Create LHE content with valid init and event but empty weights block
    empty_weights_content = """<LesHouchesEvents version="1.0">
<init>
  2212  2212  6.500000e+03  6.500000e+03  0  0  0  0  3  1
  1.000000e+00  0.000000e+00  1.000000e+00  1
</init>
<event>
  2      0 +1.0000000e+00  9.11884000e+01 -1.00000000e+00 -1.00000000e+00
       21 -1    0    0  501  502 +0.00000000e+00 +0.00000000e+00 +4.56308892e+02 +4.56308892e+02 +0.00000000e+00 0.0000e+00 9.0000e+00
       21 -1    0    0  502  501 -0.00000000e+00 -0.00000000e+00 -2.24036073e+02 +2.24036073e+02 +0.00000000e+00 0.0000e+00 9.0000e+00
<weights></weights>
</event>
</LesHouchesEvents>"""

    # Test reading events with empty weights block
    with NamedTemporaryFile(mode="w", suffix=".lhe", delete=False) as tmp_file:
        tmp_file.write(empty_weights_content)
        tmp_file_path = tmp_file.name

    try:
        with pytest.raises(ValueError, match=r"<weights> block has no text"):
            list(pylhe.read_lhe_with_attributes(tmp_file_path))
    finally:
        os.unlink(tmp_file_path)


def test_empty_wgt_block_error():
    """Test that ValueError is raised when <wgt> block has no text content."""
    # Create LHE content with valid init and event but empty wgt block
    empty_wgt_content = """<LesHouchesEvents version="1.0">
<init>
  2212  2212  6.500000e+03  6.500000e+03  0  0  0  0  3  1
  1.000000e+00  0.000000e+00  1.000000e+00  1
</init>
<event>
  2      0 +1.0000000e+00  9.11884000e+01 -1.00000000e+00 -1.00000000e+00
       21 -1    0    0  501  502 +0.00000000e+00 +0.00000000e+00 +4.56308892e+02 +4.56308892e+02 +0.00000000e+00 0.0000e+00 9.0000e+00
       21 -1    0    0  502  501 -0.00000000e+00 -0.00000000e+00 -2.24036073e+02 +2.24036073e+02 +0.00000000e+00 0.0000e+00 9.0000e+00
<rwgt>
<wgt id="1001"></wgt>
</rwgt>
</event>
</LesHouchesEvents>"""

    # Test reading events with empty wgt block
    with NamedTemporaryFile(mode="w", suffix=".lhe", delete=False) as tmp_file:
        tmp_file.write(empty_wgt_content)
        tmp_file_path = tmp_file.name

    try:
        with pytest.raises(ValueError, match=r"<wgt> block has no text"):
            list(pylhe.read_lhe_with_attributes(tmp_file_path))
    finally:
        os.unlink(tmp_file_path)


def test_whitespace_only_wgt_block_error():
    """Test that ValueError is raised when <wgt> block has only whitespace content."""
    # Create LHE content with valid init and event but whitespace-only wgt block
    whitespace_wgt_content = """<LesHouchesEvents version="1.0">
<init>
  2212  2212  6.500000e+03  6.500000e+03  0  0  0  0  3  1
  1.000000e+00  0.000000e+00  1.000000e+00  1
</init>
<event>
  2      0 +1.0000000e+00  9.11884000e+01 -1.00000000e+00 -1.00000000e+00
       21 -1    0    0  501  502 +0.00000000e+00 +0.00000000e+00 +4.56308892e+02 +4.56308892e+02 +0.00000000e+00 0.0000e+00 9.0000e+00
       21 -1    0    0  502  501 -0.00000000e+00 -0.00000000e+00 -2.24036073e+02 +2.24036073e+02 +0.00000000e+00 0.0000e+00 9.0000e+00
<rwgt>
<wgt id="1001">   </wgt>
</rwgt>
</event>
</LesHouchesEvents>"""

    # Test reading events with whitespace-only wgt block
    with NamedTemporaryFile(mode="w", suffix=".lhe", delete=False) as tmp_file:
        tmp_file.write(whitespace_wgt_content)
        tmp_file_path = tmp_file.name

    try:
        with pytest.raises(
            ValueError,
            match=r"could not convert string to float|invalid literal for float",
        ):
            list(pylhe.read_lhe_with_attributes(tmp_file_path))
    finally:
        os.unlink(tmp_file_path)



import pylhe


def test_top_level_api():
    assert dir(pylhe) == [
        "LHEEvent",
        "LHEEventInfo",
        "LHEFile",
        "LHEInit",
        "LHEInitInfo",
        "LHEParticle",
        "LHEProcInfo",
        "LHEWeightGroup",
        "LHEWeightInfo",
        "__version__",
        "read_lhe",
        "read_lhe_file",
        "read_lhe_init",
        "read_lhe_with_attributes",
        "read_num_events",
        "to_awkward",
        "write_lhe_file",
        "write_lhe_file_path",
        "write_lhe_file_string",
        "write_lhe_string",
    ]


def test_awkward_api():
    assert dir(pylhe.awkward) == ["to_awkward"]


def test_load_version():
    assert pylhe.__version__



"""
Python interface to read Les Houches Event (LHE) files.
"""

import gzip
import io
import os
import warnings
import xml.etree.ElementTree as ET
from abc import ABC
from collections.abc import Iterable, Iterator, MutableMapping
from dataclasses import asdict, dataclass, field, fields
from typing import (
    Any,
    BinaryIO,
    Optional,
    Protocol,
    TextIO,
    TypeVar,
    Union,
)

import graphviz  # type: ignore[import-untyped]
from particle import latex_to_html_name
from particle.converters.bimap import DirectionalMaps
from particle.exceptions import MatchingIDNotFound

from pylhe._version import version as __version__

__all__ = [
    "LHEEvent",
    "LHEEventInfo",
    "LHEFile",
    "LHEInit",
    "LHEInitInfo",
    "LHEParticle",
    "LHEProcInfo",
    "LHEWeightGroup",
    "LHEWeightInfo",
    "__version__",
    "read_lhe",
    "read_lhe_file",
    "read_lhe_init",
    "read_lhe_with_attributes",
    "read_num_events",
    "to_awkward",
    "write_lhe_file",
    "write_lhe_file_path",
    "write_lhe_file_string",
    "write_lhe_string",
]


def __dir__() -> list[str]:
    return __all__


# retrieve mapping of PDG ID to particle name as LaTeX string
_PDGID2LaTeXNameMap, _ = DirectionalMaps("PDGID", "LATEXNAME", converters=(str, str))

PathLike = Union[str, bytes, os.PathLike[str], os.PathLike[bytes]]


class Writeable(Protocol):
    """
    A protocol for writeable objects.
    """

    def write(self, s: str) -> Any:
        """Write a string to the object."""
        ...


TWriteable = TypeVar("TWriteable", bound=Writeable)


@dataclass
class DictCompatibility(MutableMapping[str, Any], ABC):
    """
    Mixin for dataclasses to behave like mutable dictionaries.
    """

    def __getitem__(self, key: str) -> Any:
        """
        Get a dict by fieldname.

        For backward compatibility with versions < 1.0.0.

        .. deprecated:: 1.0.0
            Access by `object['key']` is deprecated and will be removed in a future version. Use `object.key` instead.
        """
        warnings.warn(
            f'Access by `object["{key}"]` is deprecated and will be removed in a future version. '
            f"Use `object.{key}` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return getattr(self, key)

    def __setitem__(self, key: str, value: Any) -> None:
        """
        Set a dict by fieldname.

        For backward compatibility with versions < 1.0.0.

        .. deprecated:: 1.0.0
            Access by `object['key']` is deprecated and will be removed in a future version. Use `object.key` instead.
        """
        warnings.warn(
            f'Access by `object["{key}"]` is deprecated and will be removed in a future version. '
            f"Use `object.{key}` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        setattr(self, key, value)

    def __delitem__(self, key: str) -> None:
        err = f"Cannot delete field {key!r} from dataclass instance"
        raise TypeError(err)

    def __iter__(self) -> Any:
        warnings.warn(
            "Dict-like iteration is deprecated and will be removed in a future version. "
            "Use `asdict(object)` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return iter(asdict(self))

    def __len__(self) -> int:
        warnings.warn(
            "Dict-like length is deprecated and will be removed in a future version. "
            "Use `asdict(object)` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return len(asdict(self))

    @property
    def fieldnames(self) -> list[str]:
        """
        Return the fieldnames.

        For backward compatibility with versions < 1.0.0.

        .. deprecated:: 1.0.0
            Listing fieldnames via `object.fieldnames` is deprecated and will be removed in a future version.
        """
        warnings.warn(
            "The fieldnames property is deprecated and will be removed in a future version. "
            "Use `asdict(object)` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return [f.name for f in fields(self)]


@dataclass
class LHEEventInfo(DictCompatibility):
    """
    Store the event information in the LHE format.
    """

    nparticles: int
    """Number of particles in the event"""
    pid: int
    """Process ID for the event"""
    weight: float
    """Event weight"""
    scale: float
    """Energy scale of the event"""
    aqed: float
    """QED coupling constant alpha_QED"""
    aqcd: float
    """QCD coupling constant alpha_QCD"""

    def tolhe(self) -> str:
        """
        Return the event info as a string in LHE format.

        Returns:
            str: The event info as a string in LHE format.
        """
        return f"{self.nparticles:3d} {self.pid:6d} {self.weight: 15.10e} {self.scale: 15.10e} {self.aqed: 15.10e} {self.aqcd: 15.10e}"

    @classmethod
    def fromstring(cls, string: str) -> "LHEEventInfo":
        """
        Create an `LHEEventInfo` instance from a string in LHE format.
        """
        values = string.split()
        return cls(
            nparticles=int(float(values[0])),
            pid=int(float(values[1])),
            weight=float(values[2]),
            scale=float(values[3]),
            aqed=float(values[4]),
            aqcd=float(values[5]),
        )


@dataclass
class LHEParticle(DictCompatibility):
    """
    Represents a single particle in the LHE format.
    """

    id: int
    """PDG ID of the particle"""
    status: int
    """Status code of the particle"""
    mother1: int
    """First mother particle ID"""
    mother2: int
    """Second mother particle ID"""
    color1: int
    """First color line ID"""
    color2: int
    """Second color line ID"""
    px: float
    """Momentum x coordinate of the particle"""
    py: float
    """Momentum y coordinate of the particle"""
    pz: float
    """Momentum z coordinate of the particle"""
    e: float
    """Energy of the particle"""
    m: float
    """Mass of the particle"""
    lifetime: float
    """Lifetime of the particle"""
    spin: float
    """Spin of the particle"""

    def __post_init__(self) -> None:
        """Initialize the event reference."""
        # we store the circular event reference in a private attribute
        self._event: Optional[LHEEvent] = None

    @property
    def event(self) -> Optional["LHEEvent"]:
        """Reference to the parent event, set when the particle is added to an event."""
        # Previously it was just event so we still allow that for backward compatibility
        return self._event

    @event.setter
    def event(self, value: Optional["LHEEvent"]) -> None:
        """Set the parent event reference."""
        # Previously it was just event so we still allow that for backward compatibility
        self._event = value

    @classmethod
    def fromstring(cls, string: str) -> "LHEParticle":
        """
        Create an `LHEParticle` instance from a string in LHE format.
        """
        values = string.split()
        return cls(
            id=int(float(values[0])),
            status=int(float(values[1])),
            mother1=int(float(values[2])),
            mother2=int(float(values[3])),
            color1=int(float(values[4])),
            color2=int(float(values[5])),
            px=float(values[6]),
            py=float(values[7]),
            pz=float(values[8]),
            e=float(values[9]),
            m=float(values[10]),
            lifetime=float(values[11]),
            spin=float(values[12]),
        )

    def tolhe(self) -> str:
        """
        Return the particle as a string in LHE format.

        Returns:
            str: The particle as a string in LHE format.
        """
        return f"{self.id:5d} {self.status:3d} {self.mother1:3d} {self.mother2:3d} {self.color1:3d} {self.color2:3d} {self.px: 15.8e} {self.py: 15.8e} {self.pz: 15.8e} {self.e: 15.8e} {self.m: 15.8e} {self.lifetime: 10.4e} {self.spin: 10.4e}"

    def mothers(self) -> list["LHEParticle"]:
        """
        Return a list of the particle's mothers.
        """
        if self.event is None:
            err = "Particle is not associated to an event."
            raise ValueError(err)
        first_idx = int(self.mother1) - 1
        second_idx = int(self.mother2) - 1
        return [
            self.event.particles[idx] for idx in (first_idx, second_idx) if idx >= 0
        ]


def _indent(elem: ET.Element, level: int = 0) -> None:
    """
    XML indentation helper from https://stackoverflow.com/a/33956544.
    """
    i = "\n" + level * "  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for inner_elem in elem:
            _indent(inner_elem, level + 1)
        elem = inner_elem
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    elif level and (not elem.tail or not elem.tail.strip()):
        elem.tail = i


@dataclass
class LHEInitInfo(DictCompatibility):
    """Store the first line of the <init> block as a dataclass."""

    beamA: int
    """Beam A PDG ID"""
    beamB: int
    """Beam B PDG ID"""
    energyA: float
    """Beam A energy"""
    energyB: float
    """Beam B energy"""
    PDFgroupA: int
    """PDF group for beam A"""
    PDFgroupB: int
    """PDF group for beam B"""
    PDFsetA: int
    """PDF set for beam A"""
    PDFsetB: int
    """PDF set for beam B"""
    weightingStrategy: int
    """Weighting strategy"""
    numProcesses: int
    """Number of processes"""

    def tolhe(self) -> str:
        """
        Return the init info block as a string in LHE format.

        Returns:
            str: The init info block as a string in LHE format.
        """
        return f" {self.beamA: 6d} {self.beamB: 6d} {self.energyA: 14.7e} {self.energyB: 14.7e} {self.PDFgroupA: 5d} {self.PDFgroupB: 5d} {self.PDFsetA: 5d} {self.PDFsetB: 5d} {self.weightingStrategy: 5d} {self.numProcesses: 5d}"

    @classmethod
    def fromstring(cls, string: str) -> "LHEInitInfo":
        """
        Create an `LHEInitInfo` instance from a string in LHE format.
        """
        values = string.split()
        return cls(
            beamA=int(float(values[0])),
            beamB=int(float(values[1])),
            energyA=float(values[2]),
            energyB=float(values[3]),
            PDFgroupA=int(float(values[4])),
            PDFgroupB=int(float(values[5])),
            PDFsetA=int(float(values[6])),
            PDFsetB=int(float(values[7])),
            weightingStrategy=int(float(values[8])),
            numProcesses=int(float(values[9])),
        )


@dataclass
class LHEProcInfo(DictCompatibility):
    """Store the process info block as a dataclass."""

    xSection: float
    """Cross section of the process"""
    error: float
    """Uncertainty/error of the cross section"""
    unitWeight: float
    """Unit weight of the process"""
    procId: int
    """Process ID"""

    def tolhe(self) -> str:
        """
        Return the process info block as a string in LHE format.

        Returns:
            str: The process info block as a string in LHE format.
        """
        return f"{self.xSection: 14.7e} {self.error: 14.7e} {self.unitWeight: 14.7e} {self.procId: 5d}"

    @classmethod
    def fromstring(cls, string: str) -> "LHEProcInfo":
        """
        Create an `LHEProcInfo` instance from a string in LHE format.
        """
        values = string.split()
        return cls(
            xSection=float(values[0]),
            error=float(values[1]),
            unitWeight=float(values[2]),
            procId=int(float(values[3])),
        )


@dataclass
class LHEWeightInfo(DictCompatibility):
    """Information about a single weight in a weight group."""

    attrib: dict[str, str]
    """Weight XML attributes"""
    name: str
    """Weight description text"""
    index: int
    """Sequential index for ordering"""


@dataclass
class LHEWeightGroup(DictCompatibility):
    """Information about a weight group."""

    attrib: dict[str, str]
    """Weight group XML attributes"""
    weights: dict[str, LHEWeightInfo]
    """Dictionary of weight ID to weight information"""


@dataclass
class LHEInit(DictCompatibility):
    """Store the <init> block as a dataclass."""

    initInfo: LHEInitInfo
    """Init information"""
    procInfo: list[LHEProcInfo]
    """Process information"""
    weightgroup: dict[str, LHEWeightGroup]
    """Weight group information"""
    LHEVersion: str
    """LHE version"""

    def tolhe(self) -> str:
        """
        Return the init block as a string in LHE format.

        Returns:
            str: The init block as a string in LHE format.
        """
        # weightgroups to xml
        root = ET.Element("initrwgt")
        for _k, v in self.weightgroup.items():
            weightgroup_elem = ET.SubElement(root, "weightgroup", attrib=v.attrib)
            for _key, value in v.weights.items():
                weight_elem = ET.SubElement(
                    weightgroup_elem, "weight", attrib=value.attrib
                )
                weight_elem.text = value.name
        _indent(root)
        sweightgroups = ET.tostring(root, encoding="unicode", method="xml")

        return (
            "<init>\n"
            + self.initInfo.tolhe()
            + "\n"
            + "\n".join([p.tolhe() for p in self.procInfo])
            + "\n"
            + f"{sweightgroups}"
            + "</init>"
        )

    def __getitem__(self, key: str) -> Any:
        """
        Get a dict fieldname.

        For backward compatibility with versions < 1.0.0.

        .. deprecated:: 1.0.0
            Access by `lheinit["key"]` is deprecated and will be removed in a future version.
        """
        warnings.warn(
            f'Access by `lheinit["{key}"]` is deprecated and will be removed in a future version. '
            f"Use `lheinit.{key}` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        # Map field names to dataclass attributes
        if key in self.fieldnames:
            return getattr(self, key)
        # Try to get from initInfo for backward compatibility
        return getattr(self.initInfo, key)

    def __setitem__(self, key: str, value: Any) -> None:
        """
        Set a dict fieldname.

        For backward compatibility with versions < 1.0.0.

        .. deprecated:: 1.0.0
            Access by `lheinit["key"]` is deprecated and will be removed in a future version.
        """
        warnings.warn(
            f'Access by `lheinit["{key}"]` is deprecated and will be removed in a future version. '
            f"Use `lheinit.{key}` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        # Map field names to dataclass attributes
        if key in self.fieldnames:
            setattr(self, key, value)
        else:
            # Try to set on initInfo for backward compatibility
            setattr(self.initInfo, key, value)

    @classmethod
    def _fromcontext(cls, root: ET.Element, context: Any) -> "LHEInit":
        initInfo = None
        procInfo = []
        weightgroup: dict[str, LHEWeightGroup] = {}
        LHEVersion: str = ""

        if root.tag == "LesHouchesEvents":
            LHEVersion = root.attrib["version"]

        for event, element in context:
            if element.tag == "initrwgt":
                weightgroup = {}
                index = 0
                for child in element:
                    # Find all weightgroups
                    if child.tag == "weightgroup" and child.attrib != {}:
                        if "type" in child.attrib:
                            wg_type = child.attrib["type"]
                        elif "name" in child.attrib:
                            wg_type = child.attrib["name"]
                        else:
                            ae = "weightgroup must have attribute 'type' or 'name'."
                            raise AttributeError(ae)
                        _temp = LHEWeightGroup(attrib=child.attrib, weights={})
                        # Iterate over all weights in this weightgroup
                        for wc in child:
                            if wc.tag != "weight":
                                continue
                            if "id" not in wc.attrib:
                                ae = "weight must have attribute 'id'"
                                raise AttributeError(ae)
                            wg_id = wc.attrib["id"]
                            _temp.weights[wg_id] = LHEWeightInfo(
                                attrib=wc.attrib,
                                name=wc.text.strip() if wc.text else "",
                                index=index,
                            )
                            index += 1

                        weightgroup[wg_type] = _temp
            if (
                element.tag == "init" and event == "end"
            ):  # text is None before end-tag if event == "start", if there are sub-elements (e.g. MadGraph stores a <generator> tag there)
                if element.text is None:
                    err = "<init> block has no text."
                    raise ValueError(err)
                data = element.text.split("\n")[1:-1]
                initInfo = LHEInitInfo.fromstring(data[0])
                procInfo = [LHEProcInfo.fromstring(d) for d in data[1:]]

            if element.tag == "event":
                break
        if initInfo is None:
            err = "No <init> block found in the LHE file."
            raise ValueError(err)
        return LHEInit(
            initInfo=initInfo,
            procInfo=procInfo,
            weightgroup=weightgroup,
            LHEVersion=LHEVersion,
        )


@dataclass
class LHEEvent(DictCompatibility):
    """
    Store a single event in the LHE format.
    """

    eventinfo: LHEEventInfo
    """Event information"""
    particles: list[LHEParticle]
    """List of particles in the event"""
    weights: dict[str, float] = field(default_factory=dict)
    """Event weights"""
    attributes: dict[str, str] = field(default_factory=dict)
    """Event attributes"""
    optional: list[str] = field(default_factory=list)
    """Optional '#' comments stored in the event"""
    _graph: Optional[graphviz.Digraph] = None
    """Stores the graph representation of the event generated after first access of the property `lheevent.graph`"""

    def __post_init__(self) -> None:
        """Set up a bidirectional relationship between event and particles."""
        for p in self.particles:
            p.event = self

    def tolhe(self, rwgt: bool = True, weights: bool = False) -> str:
        """
        Return the event as a string in LHE format.

        Args:
            rwgt (bool): Include the weights in the 'rwgt' format.
            weights (bool): Include the weights in the 'weights' format.

        Returns:
            str: The event as a string in LHE format.
        """
        sweights = ""
        if rwgt and self.weights:
            sweights = "<rwgt>\n"
            for k, v in self.weights.items():
                sweights += f" <wgt id='{k}'>{v:11.4e}</wgt>\n"
            sweights += "</rwgt>\n"
        if weights and self.weights:
            sweights = "<weights>\n"
            for v in self.weights.values():
                sweights += f"{v:11.4e}\n"
            sweights += "</weights>\n"

        return (
            "<event>\n"
            + self.eventinfo.tolhe()
            + "\n"
            + "\n".join([p.tolhe() for p in self.particles])
            + "\n"
            + sweights
            + "</event>"
        )

    @classmethod
    def _fromcontext(
        cls,
        root: ET.Element,
        context: Any,
        lheinit: LHEInit,
        with_attributes: bool,
    ) -> Iterator["LHEEvent"]:
        index_map = _get_index_to_id_map(lheinit) if with_attributes else {}
        for event, element in context:
            if event == "end" and element.tag == "event":
                if element.text is None:
                    err = "<event> block has no text."
                    raise ValueError(err)

                data = element.text.strip().split("\n")
                eventdata_str, particles_str = data[0], data[1:]

                eventinfo = LHEEventInfo.fromstring(eventdata_str)
                particles = [
                    LHEParticle.fromstring(p)
                    for p in particles_str
                    if not p.strip().startswith("#")
                ]

                if with_attributes:
                    weights = {}
                    attrib = element.attrib
                    optional = [
                        p.strip() for p in particles_str if p.strip().startswith("#")
                    ]

                    for sub in element:
                        if sub.tag == "weights":
                            if sub.text is None:
                                err = "<weights> block has no text."
                                raise ValueError(err)
                            for i, w in enumerate(sub.text.split()):
                                if w and index_map[i] not in weights:
                                    weights[index_map[i]] = float(w)
                        elif sub.tag == "rwgt":
                            for r in sub:
                                if r.tag == "wgt":
                                    if r.text is None:
                                        err = "<wgt> block has no text."
                                        raise ValueError(err)
                                    weights[r.attrib["id"]] = float(r.text.strip())

                    yield LHEEvent(
                        eventinfo=eventinfo,
                        particles=particles,
                        weights=weights,
                        attributes=attrib,
                        optional=optional,
                    )
                else:
                    yield LHEEvent(eventinfo, particles)

                # Clear memory
                element.clear()
                root.clear()

    @property
    def graph(self) -> graphviz.Digraph:
        """
        Get the `graphviz.Digraph` object.
        The user now has full control ...

        E.g., see the source with my_LHEEvent_instance.graph.source.

        When not in notebooks the graph can easily be visualized with the
        `graphviz.Digraph.render` or `graphviz.Digraph.view` functions, e.g.:
        my_LHEEvent_instance.graph.render(filename="test", format="pdf", view=True, cleanup=True)
        """
        if self._graph is None:
            self._build_graph()
        return self._graph

    def _build_graph(self) -> None:
        """
        Navigate the particles in the event and produce a Digraph in the DOT language.
        """
        self._graph = graphviz.Digraph()
        for i, p in enumerate(self.particles):
            iid = int(p.id)
            sid = str(iid)
            try:
                name = _PDGID2LaTeXNameMap[sid]
                texlbl = f"${name}$"
                label = f'<<table border="0" cellspacing="0" cellborder="0"><tr><td>{latex_to_html_name(name)}</td></tr></table>>'
            except MatchingIDNotFound:
                texlbl = sid
                label = f'<<table border="0" cellspacing="0" cellborder="0"><tr><td>{texlbl}</td></tr></table>>'
            self._graph.node(
                str(i), label=label, attr_dict=str(p.__dict__), texlbl=texlbl
            )
        for i, p in enumerate(self.particles):
            for mom in p.mothers():
                self._graph.edge(str(self.particles.index(mom)), str(i))

    def _repr_mimebundle_(
        self,
        include: Optional[Iterable[str]] = None,
        exclude: Optional[Iterable[str]] = None,
        **kwargs: dict[str, Any],
    ) -> Any:
        """
        IPython display helper.
        """
        try:
            return self.graph._repr_mimebundle_(
                include=include, exclude=exclude, **kwargs
            )
        except AttributeError:
            return {"image/svg+xml": self.graph._repr_svg_()}  # for graphviz < 0.19


@dataclass
class LHEFile(DictCompatibility):
    """
    Represents an LHE file as a dataclass.
    """

    init: LHEInit
    """Init block"""
    events: Iterable[LHEEvent] = ()
    """Event block"""

    def write(
        self, output_stream: TWriteable, rwgt: bool = True, weights: bool = False
    ) -> TWriteable:
        """
        Write the LHE file to an output stream.
        """
        output_stream.write(f'<LesHouchesEvents version="{self.init.LHEVersion}">\n')
        output_stream.write(self.init.tolhe() + "\n")
        for e in self.events:
            output_stream.write(e.tolhe(rwgt=rwgt, weights=weights) + "\n")
        output_stream.write("</LesHouchesEvents>")
        return output_stream

    def tolhe(self, rwgt: bool = True, weights: bool = False) -> str:
        """
        Return the LHE file as a string.
        """
        return self.write(io.StringIO(), rwgt=rwgt, weights=weights).getvalue()

    def tofile(
        self,
        filepath: str,
        gz: bool = False,
        rwgt: bool = True,
        weights: bool = False,
    ) -> None:
        """
        Write the LHE file.

        Args:
            filepath: Path to the output file.
            gz: Whether to gzip the output file (ignored if filepath suffix is .gz/.gzip).
            rwgt: Whether to include weights in 'rwgt' format.
            weights: Whether to include weights in 'weights' format.
        """
        # if filepath suffix is gz, write as gz
        with _open_write_file(filepath, gz=gz) as f:
            self.write(f, rwgt=rwgt, weights=weights)

    @classmethod
    def fromstring(
        cls, string: str, with_attributes: bool = True, generator: bool = True
    ) -> "LHEFile":
        """
        Create an LHEFile instance from a string in LHE format.
        """
        return cls.frombuffer(
            io.StringIO(string), with_attributes=with_attributes, generator=generator
        )

    @classmethod
    def fromfile(
        cls, filepath: PathLike, with_attributes: bool = True, generator: bool = True
    ) -> "LHEFile":
        """
        Read an LHE file and return an LHEFile object.
        """
        return cls.frombuffer(
            _extract_fileobj(filepath),
            with_attributes=with_attributes,
            generator=generator,
        )

    @classmethod
    def frombuffer(
        cls,
        fileobject: Union[
            io.BufferedReader, gzip.GzipFile, io.StringIO, TextIO, BinaryIO
        ],
        with_attributes: bool = True,
        generator: bool = True,
    ) -> "LHEFile":
        """
        Read an LHE file and return an LHEFile object.
        """

        def _generator(lhef: LHEFile) -> Iterator[LHEEvent]:
            try:
                with fileobject as fileobj:
                    context = ET.iterparse(fileobj, events=["start", "end"])
                    _, root = next(context)  # Get the root element

                    lheinit = LHEInit._fromcontext(root, context)
                    lhef.init = lheinit
                    # First yield allows caller to advance generator to read lheinit before consuming real events
                    yield LHEEvent(
                        eventinfo=LHEEventInfo(
                            nparticles=0,
                            pid=0,
                            weight=0.0,
                            scale=0.0,
                            aqed=0.0,
                            aqcd=0.0,
                        ),
                        particles=[],
                    )
                    yield from LHEEvent._fromcontext(
                        root, context, lheinit, with_attributes
                    )

            except ET.ParseError as excep:
                print("WARNING. Parse Error:", excep)
                return

        lhef = cls(
            # dummy init, will be replaced
            init=LHEInit(
                initInfo=LHEInitInfo(0, 0, 0.0, 0.0, 0, 0, 0, 0, 0, 0),
                procInfo=[],
                weightgroup={},
                LHEVersion="3.0",
            ),
            events=[],
        )
        events = _generator(lhef)
        next(events)  # advance to read lheinit

        lhef.events = events if generator else list(events)
        return lhef


def read_lhe_file(filepath: PathLike, with_attributes: bool = True) -> LHEFile:
    """
    Read an LHE file and return an LHEFile object.

    .. deprecated:: 1.0.0
        Use `LHEFile.fromfile` instead.
    """
    warnings.warn(
        "read_lhe_file is deprecated and will be removed in a future version. "
        "Use `LHEFile.fromfile` instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return LHEFile.fromfile(filepath, with_attributes=with_attributes)


def _extract_fileobj(filepath: PathLike) -> Union[io.BufferedReader, gzip.GzipFile]:
    """
    Checks to see if a file is compressed, and if so, extract it with gzip
    so that the uncompressed file can be returned.
    It returns a file object containing XML data that will be ingested by
    ``xml.etree.ElementTree.iterparse``.

    Args:
        filepath: A path-like object or str.

    Returns:
        _io.BufferedReader or gzip.GzipFile: A file object containing XML data.
    """
    with open(filepath, "rb") as gzip_file:
        header = gzip_file.read(2)
    gzip_magic_number = b"\x1f\x8b"

    return (
        gzip.GzipFile(filepath) if header == gzip_magic_number else open(filepath, "rb")
    )


def read_lhe_init(filepath: PathLike) -> LHEInit:
    """
    Read and return the init blocks. This encodes the weight group
    and related things according to https://arxiv.org/abs/1405.1067

    Args:
        filepath: A path-like object or str.

    Returns:
        dict: Dictionary containing the init blocks of the LHE file.

    .. deprecated:: 1.0.0
        Use `LHEFile.fromfile(...).init` instead.
    """
    warnings.warn(
        "read_lhe_init is deprecated and will be removed in a future version. "
        "Use `LHEFile.fromfile(...).init` instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return LHEFile.fromfile(filepath).init


def read_lhe(filepath: PathLike) -> Iterable[LHEEvent]:
    """
    Read and yield the events in the LHE file.

    .. deprecated:: 1.0.0
        Use `LHEFile.fromfile(...).events` instead.
    """
    warnings.warn(
        "read_lhe is deprecated and will be removed in a future version. "
        "Use `LHEFile.fromfile(...).events` instead",
        DeprecationWarning,
        stacklevel=2,
    )
    yield from LHEFile.fromfile(filepath, with_attributes=False).events


def _get_index_to_id_map(init: LHEInit) -> dict[int, str]:
    """
    Produce a dictionary to map weight indices to the id of the weight.

    It is used for LHE files where there is only a list of weights per event.
    This dictionary is then used to map the list of weights to their weight id.
    Ideally, this needs to be done only once and the dictionary can be reused.

    Args:
        init (LHEInit): init block as returned by read_lhe_init

    Returns:
        dict: {weight index: weight id}
    """
    ret = {}
    for wg in init.weightgroup.values():
        for id, w in wg.weights.items():
            ret[w.index] = id
    return ret


def read_lhe_with_attributes(filepath: PathLike) -> Iterable[LHEEvent]:
    """
    Iterate through file, similar to read_lhe but also set
    weights and attributes.

    .. deprecated:: 1.0.0
        Use `LHEEvent.fromfile` with the `with_attributes` parameter instead.
    """
    warnings.warn(
        "read_lhe_with_attributes is deprecated and will be removed in a future version. "
        "Use `LHEEvent.fromfile` with the `with_attributes` parameter instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    yield from LHEFile.fromfile(filepath, with_attributes=True).events


def read_num_events(filepath: PathLike) -> int:
    """
    Moderately efficient way to get the number of events stored in a file.
    """
    try:
        with _extract_fileobj(filepath) as fileobj:
            context = ET.iterparse(fileobj, events=["start", "end"])
            _, root = next(context)  # Get the root element
            count = 0
            for event, element in context:
                if event == "end" and element.tag == "event":
                    count += 1
                    # Clear the element to free memory
                    element.clear()
                    # Root tracks sub-elements -> clear all sub-elements
                    root.clear()
            return count
    except ET.ParseError as excep:
        print("WARNING. Parse Error:", excep)
        return -1


def write_lhe_file_string(
    lhefile: LHEFile, rwgt: bool = True, weights: bool = False
) -> str:
    """
    Return the LHE file as a string.

    .. deprecated:: 1.0.0
         Use `LHEFile.tolhe` instead.
    """
    warnings.warn(
        "write_lhe_file_string is deprecated and will be removed in a future version. "
        "Use `LHEFile.tolhe` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return lhefile.tolhe(rwgt=rwgt, weights=weights)


def write_lhe_string(
    lheinit: LHEInit,
    lheevents: Iterable[LHEEvent],
    rwgt: bool = True,
    weights: bool = False,
) -> str:
    """
    Return the LHE file as a string.

    .. deprecated:: 0.9.1
       Instead of :func:`~pylhe.write_lhe_string(init,events,rwgt,weights)` use `LHEFile(init,events).tolhe(rwgt,weights)`.
    """
    warnings.warn(
        "`write_lhe_string` is deprecated and will be removed in a future version. "
        "Use `LHEFile(...).tolhe(...)` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return LHEFile(init=lheinit, events=lheevents).tolhe(rwgt=rwgt, weights=weights)


def _open_write_file(filepath: str, gz: bool = False) -> TextIO:
    if filepath.endswith((".gz", ".gzip")) or gz:
        return gzip.open(filepath, "wt")
    return open(filepath, "w")


def write_lhe_file_path(
    lhefile: LHEFile,
    filepath: str,
    gz: bool = False,
    rwgt: bool = True,
    weights: bool = False,
) -> None:
    """
    Write the LHE file.

    .. deprecated:: 1.0.0
        Use `LHEFile.tofile` instead.
    """
    warnings.warn(
        "write_lhe_file_path is deprecated and will be removed in a future version. "
        "Use `LHEFile.tofile` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    lhefile.tofile(filepath, gz=gz, rwgt=rwgt, weights=weights)


def write_lhe_file(
    lheinit: LHEInit,
    lheevents: Iterable[LHEEvent],
    filepath: str,
    gz: bool = False,
    rwgt: bool = True,
    weights: bool = False,
) -> None:
    """
    Write the LHE file.

    .. deprecated:: 0.9.1
       Instead of :func:`~pylhe.write_lhe_file(init,events,filepath,gz,rwgt,weights)` use `LHEFile(init,events).tofile(filepath,gz,rwgt,weights)`.
    """
    warnings.warn(
        "`write_lhe_file` is deprecated and will be removed in a future version. "
        "Use `LHEFile(...).tofile(...)` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    LHEFile(init=lheinit, events=lheevents).tofile(
        filepath,
        gz=gz,
        rwgt=rwgt,
        weights=weights,
    )


# we import this later to avoid circular imports
from .awkward import to_awkward  # noqa: E402



import pytest
import skhep_testdata

import pylhe

TEST_FILE_WITHOUT_WEIGHTS = skhep_testdata.data_path("pylhe-testfile-pr29.lhe")
TEST_FILE_WITH_WEIGHTS = skhep_testdata.data_path("pylhe-testlhef3.lhe")


def test_to_awkward():
    arr = pylhe.to_awkward(pylhe.read_lhe_with_attributes(TEST_FILE_WITHOUT_WEIGHTS))
    assert len(arr) == 791
    assert len(arr) == len(arr.particles)
    assert len(arr) == len(arr.eventinfo)
    for field in arr.particles.fields:
        assert len(arr) == len(arr.particles[field])
    for field in arr.eventinfo.fields:
        assert len(arr) == len(arr.eventinfo[field])
    assert "weights" not in arr.fields

    arr = pylhe.to_awkward(pylhe.read_lhe_with_attributes(TEST_FILE_WITH_WEIGHTS))
    assert len(arr) == 59
    assert len(arr) == len(arr.particles)
    assert len(arr) == len(arr.eventinfo)
    for field in arr.particles.fields:
        assert len(arr) == len(arr.particles[field])
    for field in arr.eventinfo.fields:
        assert len(arr) == len(arr.eventinfo[field])
    assert "weights" in arr.fields
    assert len(arr) == len(arr.weights)
    assert arr.weights.fields == [
        "1001",
        "1002",
        "1003",
        "1004",
        "1005",
        "1006",
        "1007",
        "1008",
        "1009",
    ]
    assert arr.weights["1001"][0] == pytest.approx(0.50109e02)
    assert arr.weights["1002"][0] == pytest.approx(0.45746e02)
    assert arr.weights["1003"][0] == pytest.approx(0.52581e02)
    assert arr.weights["1004"][0] == pytest.approx(0.50109e02)
    assert arr.weights["1005"][0] == pytest.approx(0.45746e02)
    assert arr.weights["1006"][0] == pytest.approx(0.52581e02)
    assert arr.weights["1007"][0] == pytest.approx(0.50109e02)
    assert arr.weights["1008"][0] == pytest.approx(0.45746e02)
    assert arr.weights["1009"][0] == pytest.approx(0.52581e02)
    for field in arr.weights.fields:
        assert len(arr) == len(arr.weights[field])


def test_awkward_registration():
    arr = pylhe.to_awkward(pylhe.read_lhe_with_attributes(TEST_FILE_WITHOUT_WEIGHTS))
    assert len(arr.particles.vector.mass) == len(arr.particles)

    arr = pylhe.to_awkward(pylhe.read_lhe_with_attributes(TEST_FILE_WITH_WEIGHTS))
    assert len(arr.particles.vector.mass) == len(arr.particles)
    assert len(arr.particles.vector.mass) == len(arr.weights)


def test_to_awkward_vector():
    """
    Test numeric equality of momenta represented by vectors.
    """
    arr = pylhe.to_awkward(pylhe.read_lhe_with_attributes(TEST_FILE_WITHOUT_WEIGHTS))

    assert arr.particles.vector.px[0][0] == pytest.approx(-3.1463804033e-01)
    assert arr.particles.vector.x[0][0] == pytest.approx(-3.1463804033e-01)

    assert arr.particles.vector.py[0][0] == pytest.approx(-6.3041724109e-01)
    assert arr.particles.vector.y[0][0] == pytest.approx(-6.3041724109e-01)

    assert arr.particles.vector.pz[0][0] == pytest.approx(8.5343193374e00)
    assert arr.particles.vector.z[0][0] == pytest.approx(8.5343193374e00)

    assert arr.particles.vector.e[0][0] == pytest.approx(8.5644657479e00)
    assert arr.particles.vector.t[0][0] == pytest.approx(8.5644657479e00)



import os
from tempfile import NamedTemporaryFile

import pytest

import pylhe


def test_lhe_weight_info_getitem_deprecation_warning():
    """Test that DeprecationWarning is raised when using __getitem__ on LHEWeightInfo."""
    weight_info = pylhe.LHEWeightInfo(
        attrib={"id": "test_weight"}, name="Test weight", index=0
    )

    with pytest.warns(
        DeprecationWarning,
        match=r'Access by `object\["attrib"\]` is deprecated and will be removed in a future version\. Use `object\.attrib` instead\.',
    ):
        _ = weight_info["attrib"]


def test_lhe_weight_info_setitem_deprecation_warning():
    """Test that DeprecationWarning is raised when using __setitem__ on LHEWeightInfo."""
    weight_info = pylhe.LHEWeightInfo(
        attrib={"id": "test_weight"}, name="Test weight", index=0
    )

    with pytest.warns(
        DeprecationWarning,
        match=r'Access by `object\["name"\]` is deprecated and will be removed in a future version\. Use `object\.name` instead\.',
    ):
        weight_info["name"] = "New test weight"


def test_lhe_weight_info_iter_deprecation_warning():
    """Test that DeprecationWarning is raised when using __iter__ on LHEWeightInfo."""
    weight_info = pylhe.LHEWeightInfo(
        attrib={"id": "test_weight"}, name="Test weight", index=0
    )

    with pytest.warns(
        DeprecationWarning,
        match=r"Dict-like iteration is deprecated and will be removed in a future version\. Use `asdict\(object\)` instead\.",
    ):
        _ = list(weight_info)


def test_lhe_weight_info_len_deprecation_warning():
    """Test that DeprecationWarning is raised when using __len__ on LHEWeightInfo."""
    weight_info = pylhe.LHEWeightInfo(
        attrib={"id": "test_weight"}, name="Test weight", index=0
    )

    with pytest.warns(
        DeprecationWarning,
        match=r"Dict-like length is deprecated and will be removed in a future version\. Use `asdict\(object\)` instead\.",
    ):
        _ = len(weight_info)


def test_lhe_weight_group_getitem_deprecation_warning():
    """Test that DeprecationWarning is raised when using __getitem__ on LHEWeightGroup."""
    weight_group = pylhe.LHEWeightGroup(attrib={"type": "test_group"}, weights={})

    with pytest.warns(
        DeprecationWarning,
        match=r'Access by `object\["attrib"\]` is deprecated and will be removed in a future version\. Use `object\.attrib` instead\.',
    ):
        _ = weight_group["attrib"]


def test_lhe_weight_group_setitem_deprecation_warning():
    """Test that DeprecationWarning is raised when using __setitem__ on LHEWeightGroup."""
    weight_group = pylhe.LHEWeightGroup(attrib={"type": "test_group"}, weights={})

    weight_info = pylhe.LHEWeightInfo(
        attrib={"id": "test_weight"}, name="Test weight", index=0
    )

    with pytest.warns(
        DeprecationWarning,
        match=r'Access by `object\["weights"\]` is deprecated and will be removed in a future version\. Use `object\.weights` instead\.',
    ):
        weight_group["weights"] = weight_info


def test_lhe_init_info_getitem_deprecation_warning():
    """Test that DeprecationWarning is raised when using __getitem__ on LHEInitInfo."""
    init_info = pylhe.LHEInitInfo(
        beamA=2212,
        beamB=2212,
        energyA=6500.0,
        energyB=6500.0,
        PDFgroupA=10800,
        PDFgroupB=10800,
        PDFsetA=0,
        PDFsetB=0,
        weightingStrategy=3,
        numProcesses=1,
    )

    with pytest.warns(
        DeprecationWarning,
        match=r'Access by `object\["beamA"\]` is deprecated and will be removed in a future version\. Use `object\.beamA` instead\.',
    ):
        _ = init_info["beamA"]


def test_lhe_init_info_setitem_deprecation_warning():
    """Test that DeprecationWarning is raised when using __setitem__ on LHEInitInfo."""
    init_info = pylhe.LHEInitInfo(
        beamA=2212,
        beamB=2212,
        energyA=6500.0,
        energyB=6500.0,
        PDFgroupA=10800,
        PDFgroupB=10800,
        PDFsetA=0,
        PDFsetB=0,
        weightingStrategy=3,
        numProcesses=1,
    )

    with pytest.warns(
        DeprecationWarning,
        match=r'Access by `object\["beamA"\]` is deprecated and will be removed in a future version\. Use `object\.beamA` instead\.',
    ):
        init_info["beamA"] = 11


def test_lhe_proc_info_getitem_deprecation_warning():
    """Test that DeprecationWarning is raised when using __getitem__ on LHEProcInfo."""
    proc_info = pylhe.LHEProcInfo(xSection=1.0, error=0.1, unitWeight=1.0, procId=1)

    with pytest.warns(
        DeprecationWarning,
        match=r'Access by `object\["xSection"\]` is deprecated and will be removed in a future version\. Use `object\.xSection` instead\.',
    ):
        _ = proc_info["xSection"]


def test_lhe_proc_info_setitem_deprecation_warning():
    """Test that DeprecationWarning is raised when using __setitem__ on LHEProcInfo."""
    proc_info = pylhe.LHEProcInfo(xSection=1.0, error=0.1, unitWeight=1.0, procId=1)

    with pytest.warns(
        DeprecationWarning,
        match=r'Access by `object\["xSection"\]` is deprecated and will be removed in a future version\. Use `object\.xSection` instead\.',
    ):
        proc_info["xSection"] = 2.0


def test_lhe_event_fieldnames_deprecation_warning():
    """Test that DeprecationWarning is raised when using fieldnames property on LHEEvent."""
    event_info = pylhe.LHEEventInfo(
        nparticles=1, pid=0, weight=1.0, scale=91.188, aqed=-1.0, aqcd=-1.0
    )
    particle = pylhe.LHEParticle(
        id=21,
        status=-1,
        mother1=0,
        mother2=0,
        color1=501,
        color2=502,
        px=0.0,
        py=0.0,
        pz=456.3,
        e=456.3,
        m=0.0,
        lifetime=0.0,
        spin=9.0,
    )
    event = pylhe.LHEEvent(eventinfo=event_info, particles=[particle])

    with pytest.warns(
        DeprecationWarning,
        match=r"The fieldnames property is deprecated and will be removed in a future version\. Use `asdict\(object\)` instead\.",
    ):
        _ = event.fieldnames


def test_lhe_event_iter_deprecation_warning():
    """Test that DeprecationWarning is raised when using __iter__ on LHEEvent."""
    event_info = pylhe.LHEEventInfo(
        nparticles=1, pid=0, weight=1.0, scale=91.188, aqed=-1.0, aqcd=-1.0
    )
    particle = pylhe.LHEParticle(
        id=21,
        status=-1,
        mother1=0,
        mother2=0,
        color1=501,
        color2=502,
        px=0.0,
        py=0.0,
        pz=456.3,
        e=456.3,
        m=0.0,
        lifetime=0.0,
        spin=9.0,
    )
    event = pylhe.LHEEvent(eventinfo=event_info, particles=[particle])

    with pytest.warns(
        DeprecationWarning,
        match=r"Dict-like iteration is deprecated and will be removed in a future version\. Use `asdict\(object\)` instead\.",
    ):
        _ = list(event)


def test_lhe_event_len_deprecation_warning():
    """Test that DeprecationWarning is raised when using __len__ on LHEEvent."""
    event_info = pylhe.LHEEventInfo(
        nparticles=1, pid=0, weight=1.0, scale=91.188, aqed=-1.0, aqcd=-1.0
    )
    particle = pylhe.LHEParticle(
        id=21,
        status=-1,
        mother1=0,
        mother2=0,
        color1=501,
        color2=502,
        px=0.0,
        py=0.0,
        pz=456.3,
        e=456.3,
        m=0.0,
        lifetime=0.0,
        spin=9.0,
    )
    event = pylhe.LHEEvent(eventinfo=event_info, particles=[particle])

    with pytest.warns(
        DeprecationWarning,
        match=r"Dict-like length is deprecated and will be removed in a future version\. Use `asdict\(object\)` instead\.",
    ):
        _ = len(event)


def test_lhe_init_getitem_deprecation_warning():
    """Test that DeprecationWarning is raised when using __getitem__ on LHEInit."""
    init_info = pylhe.LHEInitInfo(
        beamA=2212,
        beamB=2212,
        energyA=6500.0,
        energyB=6500.0,
        PDFgroupA=10800,
        PDFgroupB=10800,
        PDFsetA=0,
        PDFsetB=0,
        weightingStrategy=3,
        numProcesses=1,
    )

    lhe_init = pylhe.LHEInit(
        initInfo=init_info, procInfo=[], weightgroup={}, LHEVersion="1.0"
    )

    with pytest.warns(
        DeprecationWarning, match=r"Access by `lheinit\[\"initInfo\"\]` is deprecated"
    ):
        _ = lhe_init["initInfo"]


def test_lhe_init_setitem_deprecation_warning():
    """Test that DeprecationWarning is raised when using __setitem__ on LHEInit."""
    init_info = pylhe.LHEInitInfo(
        beamA=2212,
        beamB=2212,
        energyA=6500.0,
        energyB=6500.0,
        PDFgroupA=10800,
        PDFgroupB=10800,
        PDFsetA=0,
        PDFsetB=0,
        weightingStrategy=3,
        numProcesses=1,
    )

    lhe_init = pylhe.LHEInit(
        initInfo=init_info, procInfo=[], weightgroup={}, LHEVersion="1.0"
    )

    with pytest.warns(
        DeprecationWarning, match=r"Access by `lheinit\[\"LHEVersion\"\]` is deprecated"
    ):
        lhe_init["LHEVersion"] = "2.0"


def test_write_lhe_string_deprecation_warning():
    """Test that DeprecationWarning is raised when using write_lhe_string function."""
    init_info = pylhe.LHEInitInfo(
        beamA=2212,
        beamB=2212,
        energyA=6500.0,
        energyB=6500.0,
        PDFgroupA=10800,
        PDFgroupB=10800,
        PDFsetA=0,
        PDFsetB=0,
        weightingStrategy=3,
        numProcesses=1,
    )

    proc_info = pylhe.LHEProcInfo(xSection=1.0, error=0.1, unitWeight=1.0, procId=1)

    lhe_init = pylhe.LHEInit(
        initInfo=init_info, procInfo=[proc_info], weightgroup={}, LHEVersion="1.0"
    )

    event_info = pylhe.LHEEventInfo(
        nparticles=2, pid=0, weight=1.0, scale=91.188, aqed=-1.0, aqcd=-1.0
    )

    particles = [
        pylhe.LHEParticle(
            id=21,
            status=-1,
            mother1=0,
            mother2=0,
            color1=501,
            color2=502,
            px=0.0,
            py=0.0,
            pz=456.3,
            e=456.3,
            m=0.0,
            lifetime=0.0,
            spin=9.0,
        ),
        pylhe.LHEParticle(
            id=21,
            status=-1,
            mother1=0,
            mother2=0,
            color1=502,
            color2=501,
            px=0.0,
            py=0.0,
            pz=-224.0,
            e=224.0,
            m=0.0,
            lifetime=0.0,
            spin=9.0,
        ),
    ]

    events = [pylhe.LHEEvent(eventinfo=event_info, particles=particles)]

    with pytest.warns(
        DeprecationWarning,
        match=r"`write_lhe_string` is deprecated and will be removed in a future version",
    ):
        pylhe.write_lhe_string(lhe_init, events)


def test_write_lhe_file_deprecation_warning():
    """Test that DeprecationWarning is raised when using write_lhe_file function."""
    init_info = pylhe.LHEInitInfo(
        beamA=2212,
        beamB=2212,
        energyA=6500.0,
        energyB=6500.0,
        PDFgroupA=10800,
        PDFgroupB=10800,
        PDFsetA=0,
        PDFsetB=0,
        weightingStrategy=3,
        numProcesses=1,
    )

    proc_info = pylhe.LHEProcInfo(xSection=1.0, error=0.1, unitWeight=1.0, procId=1)

    lhe_init = pylhe.LHEInit(
        initInfo=init_info, procInfo=[proc_info], weightgroup={}, LHEVersion="1.0"
    )

    event_info = pylhe.LHEEventInfo(
        nparticles=2, pid=0, weight=1.0, scale=91.188, aqed=-1.0, aqcd=-1.0
    )

    particles = [
        pylhe.LHEParticle(
            id=21,
            status=-1,
            mother1=0,
            mother2=0,
            color1=501,
            color2=502,
            px=0.0,
            py=0.0,
            pz=456.3,
            e=456.3,
            m=0.0,
            lifetime=0.0,
            spin=9.0,
        )
    ]

    events = [pylhe.LHEEvent(eventinfo=event_info, particles=particles)]

    with NamedTemporaryFile(mode="w", suffix=".lhe", delete=False) as tmp_file:
        tmp_file_path = tmp_file.name

    try:
        with pytest.warns(
            DeprecationWarning,
            match=r"`write_lhe_file` is deprecated and will be removed in a future version",
        ):
            pylhe.write_lhe_file(lhe_init, events, tmp_file_path)
    finally:
        os.unlink(tmp_file_path)



import gzip
import os
import shutil
from pathlib import Path
from tempfile import NamedTemporaryFile

import pytest
import skhep_testdata

import pylhe
from pylhe import LHEEvent

TEST_FILE_LHE_v1 = skhep_testdata.data_path("pylhe-testfile-pr29.lhe")
TEST_FILE_LHE_v3 = skhep_testdata.data_path("pylhe-testlhef3.lhe")
TEST_FILE_LHE_INITRWGT_WEIGHTS = skhep_testdata.data_path(
    "pylhe-testfile-powheg-box-v2-hvq.lhe"
)
TEST_FILE_LHE_RWGT_WGT = skhep_testdata.data_path("pylhe-testfile-powheg-box-v2-W.lhe")
TEST_FILES_LHE_POWHEG = [
    skhep_testdata.data_path(f"pylhe-testfile-powheg-box-v2-{proc}.lhe")
    for proc in ["Z", "W", "Zj", "trijet", "directphoton", "hvq"]
]
TEST_FILES_LHE_MADGRAPH = [
    skhep_testdata.data_path("pylhe-testfile-madgraph-2.0.0-wbj.lhe"),
    skhep_testdata.data_path("pylhe-testfile-madgraph-2.2.1-Z-ckkwl.lhe.gz"),
    skhep_testdata.data_path("pylhe-testfile-madgraph-2.2.1-Z-fxfx.lhe.gz"),
    skhep_testdata.data_path("pylhe-testfile-madgraph-2.2.1-Z-mlm.lhe.gz"),
    skhep_testdata.data_path("pylhe-testfile-madgraph5-3.5.8-pp_to_jj.lhe.gz"),
]
TEST_FILES_LHE_PYTHIA = [
    skhep_testdata.data_path("pylhe-testfile-pythia-6.413-ttbar.lhe"),
    skhep_testdata.data_path("pylhe-testfile-pythia-8.3.14-weakbosons.lhe"),
]
TEST_FILES_LHE_SHERPA = [
    skhep_testdata.data_path("pylhe-testfile-sherpa-3.0.1-eejjj.lhe"),
]
TEST_FILES_LHE_WHIZARD = [
    skhep_testdata.data_path("pylhe-testfile-whizard-3.1.4-eeWW.lhe"),
]
TEST_FILES_LHE_GENERATORS = [
    *TEST_FILES_LHE_MADGRAPH,
    *TEST_FILES_LHE_POWHEG,
    *TEST_FILES_LHE_PYTHIA,
    *TEST_FILES_LHE_SHERPA,
    *TEST_FILES_LHE_WHIZARD,
]


@pytest.fixture(scope="session")
def testdata_gzip_file():
    test_data = skhep_testdata.data_path("pylhe-testfile-pr29.lhe")
    tmp_path = Path(NamedTemporaryFile().name)

    # create what is basically pylhe-testfile-pr29.lhe.gz
    with open(test_data, "rb") as readfile, gzip.open(tmp_path, "wb") as writefile:
        shutil.copyfileobj(readfile, writefile)
    yield tmp_path

    # teardown
    os.remove(tmp_path)


def test_gzip_open(testdata_gzip_file):
    assert pylhe._extract_fileobj(TEST_FILE_LHE_v1)
    assert pylhe._extract_fileobj(testdata_gzip_file)

    # Needs path-like object, not a fileobj
    with pytest.raises(TypeError), open(TEST_FILE_LHE_v1, "rb") as fileobj:
        pylhe._extract_fileobj(fileobj)

    with open(TEST_FILE_LHE_v1, "rb") as fileobj:
        assert isinstance(pylhe._extract_fileobj(TEST_FILE_LHE_v1), type(fileobj))
        assert isinstance(pylhe._extract_fileobj(Path(TEST_FILE_LHE_v1)), type(fileobj))
    assert isinstance(pylhe._extract_fileobj(testdata_gzip_file), gzip.GzipFile)
    assert isinstance(pylhe._extract_fileobj(Path(testdata_gzip_file)), gzip.GzipFile)


def test_read_num_events(testdata_gzip_file):
    assert pylhe.read_num_events(TEST_FILE_LHE_v1) == 791
    assert pylhe.read_num_events(TEST_FILE_LHE_v1) == pylhe.read_num_events(
        testdata_gzip_file
    )


def test_read_lhe_init_gzipped_file(testdata_gzip_file):
    assert pylhe.read_lhe_init(TEST_FILE_LHE_v1) == pylhe.read_lhe_init(
        testdata_gzip_file
    )


def test_read_lhe_init_v1():
    """
    Test method read_lhe_init() on a LesHouchesEvents version="1.0" file.
    """
    init_data = pylhe.read_lhe_init(TEST_FILE_LHE_v1)

    assert init_data["LHEVersion"] == "1.0"

    init_info = init_data["initInfo"]
    assert init_info["beamA"] == pytest.approx(1.0)
    assert init_info["beamB"] == pytest.approx(2.0)
    assert init_info["energyA"] == pytest.approx(1.234567)
    assert init_info["energyB"] == pytest.approx(2.345678)
    assert init_info["PDFgroupA"] == pytest.approx(3.0)
    assert init_info["PDFgroupB"] == pytest.approx(4.0)
    assert init_info["PDFsetA"] == pytest.approx(5.0)
    assert init_info["PDFsetB"] == pytest.approx(6.0)
    assert init_info["weightingStrategy"] == pytest.approx(7.0)
    assert init_info["numProcesses"] == pytest.approx(8.0)

    assert init_data["procInfo"] == []


def test_read_lhe_init_v3():
    """
    Test method read_lhe_init() on a LesHouchesEvents version="3.0" file.
    """
    init_data = pylhe.read_lhe_init(TEST_FILE_LHE_v3)

    assert len(init_data["weightgroup"]) == 1
    assert len(init_data["weightgroup"]["scale_variation"]["weights"]) == 9


def test_read_lhe_v1():
    """
    Test method read_lhe() on a LesHouchesEvents version="1.0" file.
    """
    events = pylhe.read_lhe(TEST_FILE_LHE_v1)

    assert events
    for e in events:
        assert isinstance(e, LHEEvent)


def test_read_lhe_v3():
    """
    Test method read_lhe() on a LesHouchesEvents version="3.0" file.
    """
    events = pylhe.read_lhe(TEST_FILE_LHE_v3)

    assert events
    for e in events:
        assert isinstance(e, LHEEvent)


def test_read_lhe_with_attributes_v1():
    """
    Test method read_lhe_with_attributes() on a LesHouchesEvents version="1.0" file.
    """
    events = pylhe.read_lhe_with_attributes(TEST_FILE_LHE_v1)

    assert events
    for e in events:
        assert isinstance(e, LHEEvent)


def test_read_lhe_with_attributes_v3():
    """
    Test method read_lhe_with_attributes() on a LesHouchesEvents version="3.0" file.
    """
    events = pylhe.read_lhe_with_attributes(TEST_FILE_LHE_v3)

    assert events
    for e in events:
        assert isinstance(e, LHEEvent)


@pytest.mark.parametrize("file", TEST_FILES_LHE_GENERATORS)
def test_read_lhe_generator(file):
    """
    Test method read_lhe() on several types of LesHouchesEvents generator files.
    """
    events = pylhe.read_lhe(file)

    assert events
    for e in events:
        assert isinstance(e, LHEEvent)


@pytest.mark.parametrize("file", TEST_FILES_LHE_GENERATORS)
def test_read_lhe_with_attributes_generator(file):
    """
    Test method read_lhe_with_attributes() on several types of LesHouchesEvents generator files.
    """
    events = pylhe.read_lhe_with_attributes(file)

    assert events
    for e in events:
        assert isinstance(e, LHEEvent)


@pytest.mark.parametrize(
    "file", [TEST_FILE_LHE_INITRWGT_WEIGHTS, TEST_FILE_LHE_RWGT_WGT]
)
def test_read_lhe_file(file):
    """
    Test that the read_lhe_file function works as the individual reads.
    """
    lhefile = pylhe.read_lhe_file(file, with_attributes=False)
    lheinit = pylhe.read_lhe_init(file)
    lheevents = pylhe.read_lhe(file)

    assert lheinit == lhefile.init
    assert next(lheevents).tolhe() == next(lhefile.events).tolhe()

    lhefile = pylhe.read_lhe_file(file, with_attributes=True)
    lheevents = pylhe.read_lhe_with_attributes(file)

    assert lheinit == lhefile.init
    assert next(lheevents).tolhe() == next(lhefile.events).tolhe()


@pytest.mark.parametrize(
    "file", [TEST_FILE_LHE_INITRWGT_WEIGHTS, TEST_FILE_LHE_RWGT_WGT]
)
def test_read_buffer(file):
    with open(file, "rb") as f:
        lhef1 = pylhe.LHEFile.frombuffer(f, with_attributes=True)
        next(lhef1.events)
    next(lhef1.events)


def test_read_lhe_initrwgt_weights():
    """
    Test the weights from initrwgt with a weights list.
    """
    events = pylhe.read_lhe_with_attributes(TEST_FILE_LHE_INITRWGT_WEIGHTS)

    assert events
    for e in events:
        assert isinstance(e, LHEEvent)
        assert len(e.weights) > 0


def test_read_lhe_rwgt_wgt():
    """
    Test the weights from rwgt with a wgt list.
    """
    events = pylhe.read_lhe_with_attributes(TEST_FILE_LHE_RWGT_WGT)

    assert events
    for e in events:
        assert isinstance(e, LHEEvent)
        assert len(e.weights) > 0


def test_issue_102():
    """
    Test a file containing lines starting with "#aMCatNLO".
    """
    assert pylhe.read_num_events(TEST_FILE_LHE_v3) == 59
    assert len(list(pylhe.read_lhe(TEST_FILE_LHE_v3))) == len(
        list(pylhe.read_lhe_with_attributes(TEST_FILE_LHE_v3))
    )


def test_read_lhe_init_raises():
    """
    Test that the <init> block raises AttributeErrors on faulty inputs.
    """

    with pytest.raises(
        AttributeError, match=r"weightgroup must have attribute 'type' or 'name'."
    ):
        pylhe.LHEFile.fromstring(
            """<init>
   2212   2212  4.0000000e+03  4.0000000e+03    -1    -1  21100  21100    -4     1
 5.0109086e+01  8.9185414e-02  5.0109093e+01    66
<initrwgt>
  <weightgroup combine="envelope">
    <weight id="1001">muR=0.10000E+01 muF=0.10000E+01</weight>
    <weight id="1002">muR=0.10000E+01 muF=0.20000E+01</weight>
    <weight id="1003">muR=0.10000E+01 muF=0.50000E+00</weight>
    <weight id="1004">muR=0.20000E+01 muF=0.10000E+01</weight>
    <weight id="1005">muR=0.20000E+01 muF=0.20000E+01</weight>
    <weight id="1006">muR=0.20000E+01 muF=0.50000E+00</weight>
    <weight id="1007">muR=0.50000E+00 muF=0.10000E+01</weight>
    <weight id="1008">muR=0.50000E+00 muF=0.20000E+01</weight>
    <weight id="1009">muR=0.50000E+00 muF=0.50000E+00</weight>
  </weightgroup>
</initrwgt>
</init>"""
        )

    with pytest.raises(AttributeError, match=r"weight must have attribute 'id'"):
        pylhe.LHEFile.fromstring(
            """<init>
   2212   2212  4.0000000e+03  4.0000000e+03    -1    -1  21100  21100    -4     1
 5.0109086e+01  8.9185414e-02  5.0109093e+01    66
<initrwgt>
  <weightgroup name="a fake name" combine="envelope">
    <spam>muR=0.10000E+01 muF=0.10000E+01</spam>
    <weight id="1001">muR=0.10000E+01 muF=0.10000E+01</weight>
    <weight id="1002">muR=0.10000E+01 muF=0.20000E+01</weight>
    <weight id="1003">muR=0.10000E+01 muF=0.50000E+00</weight>
    <weight id="1004">muR=0.20000E+01 muF=0.10000E+01</weight>
    <weight id="1005">muR=0.20000E+01 muF=0.20000E+01</weight>
    <weight id="1006">muR=0.20000E+01 muF=0.50000E+00</weight>
    <weight id="1007">muR=0.50000E+00 muF=0.10000E+01</weight>
    <weight id="1008">muR=0.50000E+00 muF=0.20000E+01</weight>
    <weight>muR=0.50000E+00 muF=0.50000E+00</weight>
  </weightgroup>
</initrwgt>
</init>"""
        )


def test_event_at_position_5():
    """
    Test that the event at position 5 has the expected values.
    The element at position 5 in the LHE file is:

    <event>
         5     0  1.554392E-03  0.000000E+00  0.000000E+00  0.000000E+00
         111    0    0    0    0    0 -9.7035523745E-01 -9.8435906372E-01  5.1424008917E+00  5.3267140888E+00  1.3800000000E-01 0. 9.
         211    0    0    0    0    0 -2.1024089632E-01 -3.2883303721E-02  3.1406432734E+00  3.1508676134E+00  1.3800000000E-01 0. 9.
        -211    0    0    0    0    0  3.9695103971E-02 -2.2872518121E-01  1.0496526207E-01  2.8974577830E-01  1.3800000000E-01 0. 9.
        2212    0    0    0    0    0 -8.6160811751E-01 -7.3819273849E-01  7.4246467306E+00  7.5762050386E+00  9.9307937557E-01 0. 9.
         211    0    0    0    0    0 -4.7889071830E-01 -2.8340027352E-01  1.5195379346E+00  1.6240971553E+00  1.3800000000E-01 0. 9.
    # 5    34  1.5543917618E-03   3.6288856778E+01  0.0000000000E+00  0.0000000000E+00  3.6288856778E+01   1.9388062018E+01  2.3242767182E+00  2.1256769904E+00  1.9130504008E+01   9.0374892691E-01 -1.4114
    </event>
    """
    events = pylhe.read_lhe(TEST_FILE_LHE_v1)

    # Get the event at position 5 (0-indexed)
    target_event = None
    for i, event in enumerate(events):
        if i == 5:
            target_event = event
            break

    assert target_event is not None, "Event at position 5 should exist"

    # Test event info values from the LHE format comment
    assert target_event.eventinfo.nparticles == pytest.approx(5.0)
    assert target_event.eventinfo.pid == pytest.approx(0.0)
    assert target_event.eventinfo.weight == pytest.approx(1.554392e-03)
    assert target_event.eventinfo.scale == pytest.approx(0.0)
    assert target_event.eventinfo.aqed == pytest.approx(0.0)
    assert target_event.eventinfo.aqcd == pytest.approx(0.0)

    # Test that we have the expected number of particles
    assert len(target_event.particles) == 5

    # Test particle properties based on the LHE event data
    # First particle: 111 (pi0)
    first_particle = target_event.particles[0]
    assert first_particle.id == pytest.approx(111.0)  # pi0
    assert first_particle.status == pytest.approx(0.0)
    assert first_particle.mother1 == pytest.approx(0.0)
    assert first_particle.mother2 == pytest.approx(0.0)
    assert first_particle.px == pytest.approx(-9.7035523745e-01)
    assert first_particle.py == pytest.approx(-9.8435906372e-01)
    assert first_particle.pz == pytest.approx(5.1424008917e00)
    assert first_particle.e == pytest.approx(5.3267140888e00)
    assert first_particle.m == pytest.approx(1.3800000000e-01)

    # Second particle: 211 (pi+)
    second_particle = target_event.particles[1]
    assert second_particle.id == pytest.approx(211.0)  # pi+
    assert second_particle.status == pytest.approx(0.0)
    assert second_particle.px == pytest.approx(-2.1024089632e-01)
    assert second_particle.py == pytest.approx(-3.2883303721e-02)
    assert second_particle.pz == pytest.approx(3.1406432734e00)
    assert second_particle.e == pytest.approx(3.1508676134e00)
    assert second_particle.m == pytest.approx(1.3800000000e-01)

    # Third particle: -211 (pi-)
    third_particle = target_event.particles[2]
    assert third_particle.id == pytest.approx(-211.0)  # pi-
    assert third_particle.status == pytest.approx(0.0)
    assert third_particle.px == pytest.approx(3.9695103971e-02)
    assert third_particle.py == pytest.approx(-2.2872518121e-01)
    assert third_particle.pz == pytest.approx(1.0496526207e-01)
    assert third_particle.e == pytest.approx(2.8974577830e-01)
    assert third_particle.m == pytest.approx(1.3800000000e-01)

    # Fourth particle: 2212 (proton)
    fourth_particle = target_event.particles[3]
    assert fourth_particle.id == pytest.approx(2212.0)  # proton
    assert fourth_particle.status == pytest.approx(0.0)
    assert fourth_particle.px == pytest.approx(-8.6160811751e-01)
    assert fourth_particle.py == pytest.approx(-7.3819273849e-01)
    assert fourth_particle.pz == pytest.approx(7.4246467306e00)
    assert fourth_particle.e == pytest.approx(7.5762050386e00)
    assert fourth_particle.m == pytest.approx(9.9307937557e-01)

    # Fifth particle: 211 (pi+)
    fifth_particle = target_event.particles[4]
    assert fifth_particle.id == pytest.approx(211.0)  # pi+
    assert fifth_particle.status == pytest.approx(0.0)
    assert fifth_particle.px == pytest.approx(-4.7889071830e-01)
    assert fifth_particle.py == pytest.approx(-2.8340027352e-01)
    assert fifth_particle.pz == pytest.approx(1.5195379346e00)
    assert fifth_particle.e == pytest.approx(1.6240971553e00)
    assert fifth_particle.m == pytest.approx(1.3800000000e-01)

    # Test that all particles have proper parent-child relationships
    for particle in target_event.particles:
        assert hasattr(particle, "event")
        assert particle.event is target_event



"""
`Awkward array <https://github.com/scikit-hep/awkward>`_ interface for `pylhe`.
"""

from collections.abc import Iterable

import awkward as ak  # type: ignore[import-untyped]
import vector

from pylhe import LHEEvent

__all__ = ["to_awkward"]


def __dir__() -> list[str]:
    return __all__


def to_awkward(event_iterable: Iterable[LHEEvent]) -> ak.Array:
    """Convert an iterable of LHEEvent instances to an Awkward-Array.

    Uses Awkward's ArrayBuilder to construct the array by iterating over the events.
    The events_iterable should yield instances of LHEEvent.
    This is typically created by one of the reading functions pylhe provides like
    pylhe.LHEFile.fromfile(filepath).events.

    Args:
        event_iterable (iterable): An iterable of LHEEvent instances.

    Returns:
        awkward.Array: An Awkward array of all the events.
    """

    builder = ak.ArrayBuilder()
    for event in event_iterable:
        with builder.record(name="Event"):
            builder.field("eventinfo")
            with builder.record(name="EventInfo"):
                for fname in event.eventinfo.fieldnames:
                    builder.field(fname).real(getattr(event.eventinfo, fname))
            if event.weights != {}:
                builder.field("weights")
                with builder.record(name="Weights"):
                    for label, w in event.weights.items():
                        builder.field(label).real(w)
            builder.field("particles")
            with builder.list():
                for particle in event.particles:
                    with builder.record(name="Particle"):
                        builder.field("vector")
                        with builder.record(name="Momentum4D"):
                            for fname in ["px", "py", "pz", "e"]:
                                builder.field(fname).real(getattr(particle, fname))
                        for fname in particle.fieldnames:
                            if fname not in ["px", "py", "pz", "e"]:
                                builder.field(fname).real(getattr(particle, fname))
    return builder.snapshot()  # awkward array


# Used to register Awkward behaviors
class Particle:
    pass


class Event:
    pass


class EventInfo:
    pass


class Weights:
    pass


# Register Awkward behaviors
# See https://awkward-array.org/doc/main/reference/generated/ak.mixin_class.html
# and https://awkward-array.org/doc/main/reference/ak.behavior.html#mixin-decorators
vector.register_awkward()
ak.mixin_class(ak.behavior)(Particle)
ak.mixin_class(ak.behavior)(Event)
ak.mixin_class(ak.behavior)(EventInfo)
ak.mixin_class(ak.behavior)(Weights)
